import asyncio

from aiohttp import client_exceptions

from .config import Config
from .groups import Groups
from .lights import Lights
from .scenes import Scenes
from .errors import raise_error, ResponseError, RequestError


class Bridge:
    """Control a Hue bridge."""

    def __init__(self, host, websession, *, username=None):
        self.host = host
        self.username = username
        self.websession = websession
        self.username = username

        self.config = None
        self.groups = None
        self.lights = None
        self.scenes = None

        # self.capabilities = None
        # self.rules = None
        # self.schedules = None
        # self.sensors = None

    async def create_user(self, device_type):
        """Create a user.

        https://developers.meethue.com/documentation/configuration-api#71_create_user
        """
        result = await self.request('post', '', {
            'devicetype': device_type
        }, auth=False)
        self.username = result[0]['success']['username']
        return self.username

    async def initialize(self):
        try:
            result = await self.request('get', '')
        except client_exceptions.ClientError:
            raise RequestError(
                'Unable to connect to {}'.format(self.host)) from None

        self.config = Config(result['config'], self.request)
        self.groups = Groups(result['groups'], self.request)
        self.lights = Lights(result['lights'], self.request)
        self.scenes = Scenes(result['scenes'], self.request)

    async def request(self, method, path, json=None, auth=True):
        """Make a request to the API."""
        url = 'http://{}/api/'.format(self.host)
        if auth:
            url += '{}/'.format(self.username)
        url += path

        async with self.websession.request(method, url, json=json) as res:
            if res.content_type != 'application/json':
                raise ResponseError(
                    'Invalid content type: {}'.format(res.content_type))
            data = await res.json()
            _raise_on_error(data)
            return data


def _raise_on_error(data):
    """Check response for error message."""
    if isinstance(data, list):
        data = data[0]

    if isinstance(data, dict) and 'error' in data:
        raise_error(data['error'])



from .api import APIItems


class Groups(APIItems):
    """Represents Hue Groups.

    https://developers.meethue.com/documentation/groups-api
    """

    def __init__(self, raw, request):
        super().__init__(raw, request, 'groups', Group)

    async def get_all_lights_group(self):
        """Special all lights group."""
        return Group('0', await self._request('get', 'groups/0'), self._request)


class Group:
    """Represents a Hue Group."""
    def __init__(self, id, raw, request):
        self.id = id
        self.raw = raw
        self._request = request

    @property
    def name(self):
        return self.raw['name']

    @property
    def action(self):
        return self.raw['action']

    @property
    def state(self):
        return self.raw['state']

    @property
    def type(self):
        return self.raw['type']

    async def set_action(self, on=None, bri=None, hue=None, sat=None, xy=None,
                         ct=None, alert=None, effect=None, transitiontime=None,
                         bri_inc=None, sat_inc=None, hue_inc=None, ct_inc=None,
                         xy_inc=None, scene=None):
        """Change action of a group."""
        data = {
            key: value for key, value in {
                'on': on,
                'bri': bri,
                'hue': hue,
                'sat': sat,
                'xy': xy,
                'ct': ct,
                'alert': alert,
                'effect': effect,
                'transitiontime': transitiontime,
                'bri_inc': bri_inc,
                'sat_inc': sat_inc,
                'hue_inc': hue_inc,
                'ct_inc': ct_inc,
                'xy_inc': xy_inc,
                'scene': scene,
            }.items() if value is not None
        }

        await self._request('put', 'groups/{}/action'.format(self.id),
                            json=data)



"""Aiohue errors.

https://developers.meethue.com/documentation/error-messages
"""


class AiohueException(Exception):
    """Base error for aiohue."""


class RequestError(AiohueException):
    """Unable to fulfill request.

    Raised when host or API cannot be reached.
    """


class ResponseError(AiohueException):
    """Invalid response."""


class Unauthorized(AiohueException):
    """Username is not authorized."""


class LinkButtonNotPressed(AiohueException):
    """Raised when trying to create a user but link button not pressed."""


ERRORS = {
    1: Unauthorized,
    101: LinkButtonNotPressed
}


def raise_error(error):
    type = error['type']
    cls = ERRORS.get(type, AiohueException)
    raise cls("{}: {}".format(type, error['description']))



from .bridge import Bridge

URL_NUPNP = 'https://www.meethue.com/api/nupnp'


async def discover_nupnp(websession):
    """Discover bridges via NUPNP."""
    async with websession.get(URL_NUPNP) as res:
        return [Bridge(item['internalipaddress'], websession=websession)
                for item in (await res.json())]



from .api import APIItems


class Scenes(APIItems):
    """Represents Hue Scenes.

    https://developers.meethue.com/documentation/scenes-api
    """

    def __init__(self, raw, request):
        super().__init__(raw, request, 'scenes', Scene)


class Scene:
    """Represents a Hue Scene."""
    def __init__(self, id, raw, request):
        self.id = id
        self.raw = raw
        self._request = request

    @property
    def name(self):
        return self.raw['name']



from setuptools import setup, find_packages


long_description = open('README.md').read()

setup(
    name='aiohue',
    version='1.1.0',
    license='Apache License 2.0',
    url='https://github.com/balloob/aiohue',
    author='Paulus Schoutsen',
    author_email='paulus@paulusschoutsen.nl',
    description='Python module to talk to Philips Hue.',
    long_description=long_description,
    packages=['aiohue'],
    zip_safe=True,
    platforms='any',
    install_requires=list(val.strip() for val in open('requirements.txt')),
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



class Config:
    """Represent Hue config.

    https://developers.meethue.com/documentation/configuration-api#72_get_configuration
    """

    def __init__(self, raw, request):
        self.raw = raw
        self._request = request

    @property
    def name(self):
        """Name of the bridge."""
        return self.raw['name']

    @property
    def bridgeid(self):
        """ID of the bridge."""
        return self.raw['bridgeid']

    @property
    def mac(self):
        """Mac address of the bridge."""
        return self.raw['mac']

    async def update(self):
        self.raw = await self._request('get', 'config')
